@@ -73,6 +73,56 @@ MAX_UPLOAD_PARTS = 10_000  # maximum number of parts for S3 multipart upload
 if ClientPayloadError is not None:
     S3_RETRYABLE_ERRORS += (ClientPayloadError,)
 
+
+def add_retryable_error(exc):
+    """
+    Add an exception type to the list of retryable S3 errors.
+
+    Parameters
+    ----------
+    exc : Exception
+        The exception type to add to the retryable errors.
+
+    Examples
+    ----------
+    >>> class MyCustomError(Exception):  # doctest: +SKIP
+    ...     pass  # doctest: +SKIP
+    >>> add_retryable_error(MyCustomError)  # doctest: +SKIP
+    """
+    global S3_RETRYABLE_ERRORS
+    S3_RETRYABLE_ERRORS += (exc,)
+
+
+CUSTOM_ERROR_HANDLER = lambda _: False
+
+
+def set_custom_error_handler(func):
+    """Set a custom error handler function for S3 retryable errors.
+
+    The function should take an exception instance as its only argument,
+    and return True if the operation should be retried, or False otherwise.
+    This can also be used for custom behavior on `ClientError` exceptions,
+    such as retrying other patterns.
+
+    Parameters
+    ----------
+    func : callable[[Exception], bool]
+        The custom error handler function.
+
+    Examples
+    ----------
+    >>> def my_handler(e):  # doctest: +SKIP
+    ...     return isinstance(e, MyCustomError) and "some condition" in str(e)  # doctest: +SKIP
+    >>> set_custom_error_handler(my_handler)  # doctest: +SKIP
+
+    >>> def another_handler(e):  # doctest: +SKIP
+    ...     return isinstance(e, ClientError) and "Throttling" in str(e)"  # doctest: +SKIP
+    >>> set_custom_error_handler(another_handler)  # doctest: +SKIP
+    """
+    global CUSTOM_ERROR_HANDLER
+    CUSTOM_ERROR_HANDLER = func
+
+
 _VALID_FILE_MODES = {"r", "w", "a", "rb", "wb", "ab"}
 
 _PRESERVE_KWARGS = [
@@ -110,29 +160,46 @@ buck_acls = {"private", "public-read", "public-read-write", "authenticated-read"
 async def _error_wrapper(func, *, args=(), kwargs=None, retries):
     if kwargs is None:
         kwargs = {}
+    err = None
     for i in range(retries):
+        wait_time = min(1.7**i * 0.1, 15)
+
         try:
             return await func(*args, **kwargs)
         except S3_RETRYABLE_ERRORS as e:
             err = e
             logger.debug("Retryable error: %s", e)
-            await asyncio.sleep(min(1.7**i * 0.1, 15))
+            await asyncio.sleep(wait_time)
         except ClientError as e:
             logger.debug("Client error (maybe retryable): %s", e)
             err = e
-            wait_time = min(1.7**i * 0.1, 15)
-            if "SlowDown" in str(e):
-                await asyncio.sleep(wait_time)
-            elif "reduce your request rate" in str(e):
-                await asyncio.sleep(wait_time)
-            elif "XAmzContentSHA256Mismatch" in str(e):
+
+            matched = False
+            for pattern in [
+                "SlowDown",
+                "reduce your request rate",
+                "XAmzContentSHA256Mismatch",
+            ]:
+                if pattern in str(e):
+                    matched = True
+                    break
+
+            if matched:
                 await asyncio.sleep(wait_time)
             else:
-                break
+                should_retry = CUSTOM_ERROR_HANDLER(e)
+                if should_retry:
+                    await asyncio.sleep(wait_time)
+                else:
+                    break
         except Exception as e:
-            logger.debug("Nonretryable error: %s", e)
             err = e
-            break
+            should_retry = CUSTOM_ERROR_HANDLER(e)
+            if should_retry:
+                await asyncio.sleep(wait_time)
+            else:
+                logger.debug("Nonretryable error: %s", e)
+                break
 
     if "'coroutine'" in str(err):
         # aiobotocore internal error - fetch original botocore error
@@ -470,6 +537,7 @@ class S3FileSystem(AsyncFileSystem):
         >>> split_path("s3://mybucket/path/to/versioned_file?versionId=some_version_id")
         ['mybucket', 'path/to/versioned_file', 'some_version_id']
         """
+        trail = path[len(path.rstrip("/")) :]
         path = self._strip_protocol(path)
         path = path.lstrip("/")
         if "/" not in path:
@@ -477,6 +545,7 @@ class S3FileSystem(AsyncFileSystem):
         else:
             bucket, keypart = self._find_bucket_key(path)
             key, _, version_id = keypart.partition("?versionId=")
+            key += trail  # restore trailing slashes removed by AbstractFileSystem._strip_protocol
             return (
                 bucket,
                 key,
@@ -519,7 +588,12 @@ class S3FileSystem(AsyncFileSystem):
         >>> s3.connect(refresh=True)  # doctest: +SKIP
         """
         if self._s3 is not None and not refresh:
-            return self._s3
+            hsess = getattr(getattr(self._s3, "_endpoint", None), "http_session", None)
+            if hsess is not None:
+                if all(_.closed for _ in hsess._sessions.values()):
+                    refresh = True
+            if not refresh:
+                return self._s3
         logger.debug("Setting up s3fs instance")
 
         client_kwargs = self.client_kwargs.copy()
