@@ -3068,3 +3068,31 @@ def test_find_missing_ls(s3):
     listed_no_cache = s3_no_cache.ls(BASE, detail=False)
 
     assert set(listed_cached) == set(listed_no_cache)
+
+
+def test_session_close():
+    async def run_program(run):
+        s3 = s3fs.S3FileSystem(anon=True, asynchronous=True)
+        session = await s3.set_session()
+        files = await s3._ls(
+            "s3://noaa-hrrr-bdp-pds/hrrr.20140730/conus/"
+        )  # Random open data store
+        print(f"Number of files {len(files)}")
+        await session.close()
+
+    import aiobotocore.httpsession
+
+    aiobotocore.httpsession.AIOHTTPSession
+    asyncio.run(run_program(True))
+    asyncio.run(run_program(False))
+
+
+def test_rm_recursive_prfix(s3):
+    prefix = "logs/"  # must end with "/"
+
+    # Create empty "directory" in S3
+    client = get_boto3_client()
+    client.put_object(Bucket=test_bucket_name, Key=prefix, Body=b"")
+    logs_path = f"s3://{test_bucket_name}/{prefix}"
+    s3.rm(logs_path, recursive=True)
+    assert not s3.isdir(logs_path)
